// #name = Database Fix
// #version = 1.0.0
// #date = 2004-09-09
// #author = Ben Winslow
// #warrior = Raineth
// #description = Fixes a major bug in the T2 database code that's notorious for causing tmail issues.
// #status = Stable
// #readme = scripts/autoexec/rain/txt/dbfix.txt
// #category = Rain's scripts

package dbfix {
	// returns str[%pos .. end]
	function strRest(%buf, %pos)
	{
		return getSubStr(%buf, %pos, strlen(%buf) - %pos);
	}
	
	// find the row delimiter (\S) while double-checking for an
	// escaped \ (so that \\S doesn't trigger it)
	function FindRowDelimiter(%buf, %offset)
	{
		// first, deal with any initial offsets
		%len = strlen(%buf);
		if (%offset > %len || %offset < 0)
			return -1;
		%buf = strRest(%buf, %offset);

		while (1) {
			// search for '\S' as a row delimiter...  if it's not found or
			// it's right at the start of the string, we're done.
			%hunt = strpos(%buf, "\\S");
			if (%hunt == -1)
				return %hunt;
			else if (%hunt == 0)
				return %offset;

			// otherwise, ensure that this isn't a false alarm (\\S)
			// this is ugly since we walk backwards through '\'s to see
			// whether *this* one is escaped or not
			%falsealarm = 1;
			for (%backwards = %hunt - 1; %backwards > 0; %backwards--) {
				%falsealarm = (%falsealarm + 1) & 1;
				if ((%test = getSubStr(%buf, %backwards, 1)) !$= "\\")
					break;
			}
			if (%falsealarm) {
				// if it is, skip past it
				%buf = strRest(%buf, %hunt + 2);
				%offset += %hunt + 2;
			} else {
				// if it's not, we're happy
				return %hunt + %offset;
			}
		}
	}
	
	// quick and dirty debugging
//	function dbdebug(%str)
//	{
//		new fileObject("dbdebug");
//		
//		dbdebug.openForAppend("dbdebug.txt");
//		dbdebug.writeLine(%str);
//		dbdebug.close();
//		dbdebug.delete();
//	}

	// This is a fixed version of HDPR -- the primary difference is that it
	// uses FindRowDelimiter(), above, so that the rows are split up
	// properly.  I pretty much rewrote the function, though, using the old
	// one as a model (to clean it up a bit.)
	function HandleDatabaseProxyResponse(%prefix, %params)
	{
		%id = getWord(%params, 0);

		for (%qc = 0; %qc < $DBQueryCount; %qc++)
			if (getWord($DBQueries[%qc], 0) == %id)
				break;

		if (%qc == $DBQueryCount) {
			warn("Invalid database proxy message id: " @ %id);
			return;
		}

		%lastPacket = getWord(%params, 1);

		if (!(%start = strpos(%params, ":") + 1)) {
			warn("Invalid database proxy message: " @ %params);
			return;
		}

		%data = strRest(%params, %start);

//		dbdebug(">> Data line (ID: " @ %id @ ", final: " @ (%lastPacket ? "true" : "false") @ "): " @ %data);

		// unlike the stock tribes2 code, we don't remove escapes
		// immediately (because we won't be able to differentiate \S in the
		// data from \S as a row delimiter.) Instead, we'll remove the
		// escapes just before calling the proxy function.

		// append this data to any previous data we might have with this ID
		%newdata = $DBQueryText[%qc] @ %data;
//		dbdebug(">> Accumulated data: " @ %newdata);

		%headerDone = getWord($DBQueries[%qc], 1);
		%proxyObject = getWord($DBQueries[%qc], 2);
		%proxyKey = getWord($DBQueries[%qc], 3);

		// The first two rows are always the result status and result string,
		// which are passed to onDatabaseQueryResult as soon as we get them.
		if (!%headerDone) {
			// go hunting...
			if ((%statusend = FindRowDelimiter(%newdata, 0)) != -1)
				%stringend = FindRowDelimiter(%newdata, %statusend + 2);

			// have both?
			if (%statusend != -1 && %stringend != -1) {
				// mark that we've snagged the header
				$DBQueries[%qc] = setWord($DBQueries[%qc], 1, "1");
				%headerDone = 1;
				
				// pull out the rows
				%status = getSubStr(%newdata, 0, %statusend);
				%stringstart = %statusend + 2;
				%string = getSubStr(%newdata, %stringstart, %stringend - %stringstart);
				
				// remove escapes
				%status = strreplace(%status, "\\n", "\n");
				%status = strreplace(%status, "\\\\", "\\");
				%string = strreplace(%string, "\\n", "\n");
				%string = strreplace(%string, "\\\\", "\\");

//				dbdebug(">> Result status='" @ %status @ "' string='" @ %string @ "'");
				
				// and let the proxy object handle it
				if (%proxyObject !$= "")
					%proxyObject.onDatabaseQueryResult(%status, %string, %proxyKey);
				
				// skip past these rows
				%newdata = strRest(%newdata, %stringend + 2);
			}
		}

		if (%headerDone) {
			// start handing rows out rows to the proxy object
			while ((%end = FindRowDelimiter(%newdata, 0)) != -1) {
				// grab the row
				%rowdata = getSubStr(%newdata, 0, %end);
				// skip past it
				%newdata = strRest(%newdata, %end + 2);
				
//				dbdebug("> ROW[" @ %rowdata @ "]");
				
				// remove escapes
				%rowdata = strreplace(%rowdata, "\\n", "\n");
				%rowdata = strreplace(%rowdata, "\\\\", "\\");

				// let the proxy object handle it
				if (%proxyObject !$= "")
					%proxyObject.onDatabaseRow(%rowdata, %newdata $= "" && %lastPacket, %proxyKey);
			}
		}

		if (!%lastPacket) {
			// save anything we haven't been able to process yet
			$DBQueryText[%qc] = %newdata;
		} else {
			// all done, whine if necessary and then remove the query from
			// the array
			if (!%headerDone)
				warn("Error in database query response - not enough data.");

			for (%x = %qc; %x < $DBQueryCount; %x++) {
				$DBQueries[%x] = $DBQueries[%x + 1];
				$DBQueryText[%x] = $DBQueryText[%x + 1];
			}
			$DBQueries[%x] = $DBQueryText = "";
			$DBQueryCount--;
		}
	}
};

activatePackage(dbfix);
